#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/Xrender.h>

#include <string.h>

static void set_fixed_size(Display *dpy, Window wnd,
			   unsigned int width, unsigned int height)
{
	XSizeHints hints;

	hints.flags = PSize | PMinSize | PMaxSize;
	hints.min_width = hints.max_width = hints.base_width = width;
	hints.min_height = hints.max_height = hints.base_height = height;

	XSetWMNormalHints(dpy, wnd, &hints);
}

int main(void)
{
	int run = 1, status = -1, evbase, errbase;
	unsigned int width = 640, height = 480;
	Window wnd, wnd1, wnd2, wnd3;
	XRenderPictFormat* format;
	Atom atom_wm_delete;
	Picture wndpic;
	Display* dpy;
	Picture pic;
	XEvent e;

	dpy = XOpenDisplay(0);
	if (!dpy)
		return -1;

	if (!XCompositeQueryExtension(dpy, &evbase, &errbase))
		goto outdpy;

	atom_wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", True);

	/* create window */
	wnd = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy),
				  0, 0, width, height, 0, 0, 0);

	if (!wnd)
		goto outdpy;

	format = XRenderFindStandardFormat(dpy, PictStandardRGB24);
	pic = XRenderCreatePicture(dpy, wnd, format, 0, NULL);

	if (!pic)
		goto outpic;

	set_fixed_size(dpy, wnd, width, height);
	XSelectInput(dpy, wnd, ExposureMask | StructureNotifyMask);
	XSetWMProtocols(dpy, wnd, &atom_wm_delete, 1);
	XFlush(dpy);

	XStoreName(dpy, wnd, "X11 test");
	XMapWindow(dpy, wnd);

	/* create child window */
	wnd1 = XCreateSimpleWindow(dpy, wnd, 10, 10, width / 2, height / 2,
				   0, 0, 0xFFFFFFFF);

	if (!wnd1)
		goto outwnd;

	XSelectInput(dpy, wnd1, ExposureMask | StructureNotifyMask);
	XMapWindow(dpy, wnd1);
	XFlush(dpy);

	XCompositeRedirectWindow(dpy, wnd1, CompositeRedirectManual);
	wndpic = XRenderCreatePicture(dpy, wnd1, format, 0, NULL);

	if (!wndpic)
		goto outpic1;

	/* create children inside child */
	wnd2 = XCreateSimpleWindow(dpy, wnd1, 10, 10,
				   width / 8, height / 8, 0, 0,
				   0xFFFF0000);

	if (!wnd2)
		goto outwnd1;

	XSelectInput(dpy, wnd2, ExposureMask | StructureNotifyMask);
	XMapWindow(dpy, wnd2);
	XFlush(dpy);

	wnd3 = XCreateSimpleWindow(dpy, wnd1, 30, 30,
				   width / 8, height / 8, 0, 0,
				   0xFF00FF00);

	if (!wnd3)
		goto outwnd2;

	XSelectInput(dpy, wnd3, ExposureMask | StructureNotifyMask);
	XMapWindow(dpy, wnd3);
	XFlush(dpy);

	/* event loop */
	while (run) {
		XNextEvent(dpy, &e);

		switch (e.type) {
		case Expose:
			XClearWindow(dpy, e.xany.window);

			if (e.xany.window == wnd || e.xany.window == wnd1) {
				XRenderComposite(dpy, PictOpOver, wndpic, 0, pic,
						 0, 0, 0, 0,
						 10, 10, width / 2, height / 2);
			}

			XFlush(dpy);
			break;
		case ClientMessage:
			if (e.xclient.data.l[0] == (long)atom_wm_delete) {
				XUnmapWindow(dpy, wnd);
				run = 0;
			}
			break;
		}
	}

	status = 0;

	/* cleanup */
	XDestroyWindow(dpy, wnd3);
outwnd2:
	XDestroyWindow(dpy, wnd2);
outwnd1:
	XRenderFreePicture(dpy, wndpic);
outpic1:
	XDestroyWindow(dpy, wnd1);
outwnd:
	XRenderFreePicture(dpy, pic);
outpic:
	XDestroyWindow(dpy, wnd);
outdpy:
	XCloseDisplay(dpy);
	return status;
}
